/*      $Id: i855vidctl.c,v 1.1 2004/08/06 10:55:51 damien Exp $  */

/*-
 * Copyright (c) 2004
 *      Damien Bergamini <damien.bergamini@free.fr>. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*-
 * This tool is roughly based on the 855resolution tool written by Alain 
 * Poirier which work under Linux.
 *
 * This tool allows you to change the video modes listed in your BIOS.
 * I use it on my Acer(R) TravelMate 661LCI notebook to add support for
 * 1400x1050 mode.
 *
 * http://damien.bergamini.free.fr/i855vidctl/
 */

#include <sys/cdefs.h>
#if defined(__RCSID)
__RCSID("$Id: i855vidctl.c,v 1.1 2004/08/06 10:55:51 damien Exp $");
#else
#ifndef lint
static char rcsid[] = "$Id: i855vidctl.c,v 1.1 2004/08/06 10:55:51 damien Exp $";
#endif /* not lint */
#endif

#include <sys/param.h>
#include <sys/mman.h>

#include <machine/bus.h>
#include <machine/sysarch.h>

#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>

#define VIDEO_BIOS_START	0x000c0000
#define VIDEO_BIOS_LENGTH	0x10000	/* 64KB */

static const int freq[] = { 60, 75, 85 };

#define NB_FREQ (sizeof freq / sizeof freq[0])

struct mode {
	u_int8_t	code;
	u_int8_t	depth;
	u_int16_t	resolution; /* address of resolution structure */
	u_int8_t	unused;
} __packed;

#define MODE2RES(mem, m) \
	((struct resolution *)((unsigned char *)(mem) + (m)->resolution))

struct modeline {
	u_int32_t	clock;
	
	u_int16_t	x1;
	u_int16_t	htotal;
	u_int16_t	x2;
	u_int16_t	hblank;
	u_int16_t	hsyncstart;
	u_int16_t	hsyncend;
	
	u_int16_t	y1;
	u_int16_t	vtotal;
	u_int16_t	y2;
	u_int16_t	vblank;
	u_int16_t	vsyncstart;
	u_int16_t	vsyncend;
} __packed;

struct resolution {
 	u_int8_t	unused[6];
	struct modeline	modelines[NB_FREQ];
} __packed;

static void usage(void);
static void lock(int);
static struct mode *find_modes(void *);
static void display_modes(void *, struct mode *);
static int set_mode(void *, struct mode *, int, int, int);

int
main(int argc, char **argv)
{
	int fd;
	void *mem;
	struct mode *modes;
	int code, width, height;

	if ((fd = open("/dev/mem", O_RDWR)) == -1)
		err(EX_OSFILE, "Can't open /dev/mem");

	mem = mmap((void *)VIDEO_BIOS_START, VIDEO_BIOS_LENGTH, 
	    PROT_READ | PROT_WRITE, MAP_FIXED | MAP_SHARED, fd, 
	    VIDEO_BIOS_START);
	if (mem == NULL)
		err(EX_OSERR, "Can't map video bios");

	if ((modes = find_modes(mem)) == NULL)
		errx(EX_DATAERR, "Can't find video modes table");

	if (argc == 1) {
		display_modes(mem, modes);

	} else if (argc == 3) {
		if (sscanf(argv[1], "%x", &code) != 1)
			usage();

		if (sscanf(argv[2], "%dx%d", &width, &height) != 2)
			usage();

		if (set_mode(mem, modes, code, width, height) != 0)
			errx(EX_DATAERR, "Can't find mode 0x%x", code);

	} else
		usage();

	return 0;
}

static void
usage(void)
{
	extern char *__progname;
	
	fprintf(stderr, "usage:  %s\n"
	    "\t%s mode widthxheight\n", __progname, __progname);
	
	exit(EX_USAGE);
}

static void
lock(int lock)
{
	outl(0xcf8, 0x8000005a);
	outb(0xcfe, lock ? 0x11 : 0x33);
}

static struct mode *
find_modes(void *mem)
{
	unsigned char *p = mem;
	unsigned char *end = p + VIDEO_BIOS_LENGTH - 3 * sizeof (struct mode);
	struct mode *m;

	/* 
	 * we suppose that we have found the video modes table once we get 4 
	 * consecutives "modes" (blocks begining with 0x3?)
	 */

	for (; p < end; p++) {
		m = (struct mode *)p;
		
		if (((m[0].code & 0xf0) == 0x30) &&
		    ((m[1].code & 0xf0) == 0x30) &&
		    ((m[2].code & 0xf0) == 0x30) &&
		    ((m[3].code & 0xf0) == 0x30))
			return m;
	}

	return NULL;
}

static void
display_modes(void *mem, struct mode *m)
{
	struct resolution *r;
#ifdef notdef		    
	struct modeline *l;
	int i;
#endif

	printf("List of available video modes:\n");
	for (; m->code != 0xff; m++) {
		if (m->code == 0)
			continue;
			
		r = MODE2RES(mem, m);
	
		printf("mode %02x: %2dbits %dx%d\n", m->code, m->depth, 
		    r->modelines[0].x1 + 1, r->modelines[0].y1 + 1);
#ifdef notdef		    
		for (i = 0, l = r->modelines; i < NB_FREQ; i++, l++) {
			if (l->x1 != r->modelines[0].x1)
				continue;
		
			printf("\t(%d %d %d %d %d %d %d %d %d)\n", 
			    l->clock, 
			    l->x1 + 1, l->hsyncstart, l->hsyncend, l->htotal,
			    l->y1 + 1, l->vsyncstart, l->vsyncend, l->vtotal);
		}
#endif
	}
}

static int
set_mode(void *mem, struct mode *m, int code, int width, int height)
{
	struct resolution *r;
	struct modeline *l;
	int i;

#ifdef __FreeBSD__
	if (open("/dev/io", O_RDWR) == -1)
		err(EX_OSFILE, "Can't set ioperm");

#ifdef notdef
	if (i386_set_ioperm(0xcf8, 4, 1) == -1)
		err(EX_OSERR, "Can't get ioperm");

	if (i386_set_ioperm(0xcfe, 1, 1) == -1)
		err(EX_OSERR, "Can't set ioperm");
#endif
#else
	if (i386_iopl(1) == -1)
		err(EX_OSERR, "Can't set ioperm");
#endif

	for (; m->code != 0xff; m++) {
		if (m->code != code)
			continue;

		lock(0);
			
		r = MODE2RES(mem, m);

		for (i = 0, l = r->modelines; i < NB_FREQ; i++, l++) {
			if (l->x1 != r->modelines[0].x1)
				continue;
		
			l->x1 = l->x2 = width - 1;
			l->y1 = l->y2 = height - 1;

			l->clock = freq[i] * width;
			l->hsyncstart = width + 10;
			l->hsyncend = width + 20;
			l->htotal = width + 30;
			l->vsyncstart = height + 10;
			l->vsyncend = height + 20;
			l->vtotal = height + 30;
		}

		lock(1);

		return 0;
	}

	return 1;
}
